import pytest

from pandevice.device import Vsys
from pandevice.firewall import Firewall
from pandevice.network import EthernetInterface
from pandevice.network import VirtualRouter
from pandevice.network import VirtualWire
from pandevice.network import Vlan
from pandevice.network import Zone
from pandevice.panorama import Panorama
from pandevice.panorama import Template

from tests.live import testlib


class VsysImport(object):
    CLASS = None
    VSYS_PARAM = None

    def sanity(self, dev, state_map, fw_test=False):
        state = state_map.setdefault(dev)
        if state.err:
            state.fail_func('prereq failed')

        if fw_test:
            if not isinstance(dev, Firewall):
                pytest.skip('Skipping firewall-only test')
            dev.removeall()
            dev.add(state.obj)
            vsys_list = Vsys.refreshall(dev, name_only=True)
            for v in vsys_list:
                v.refresh_variable(self.VSYS_PARAM)

        return state

    def assert_imported_into(self, vsys, state):
        found = False

        vsys_list = Vsys.refreshall(state.parent, add=False, name_only=True)
        for v in vsys_list:
            v.refresh_variable(self.VSYS_PARAM)
            if getattr(v, self.VSYS_PARAM) is None:
                setattr(v, self.VSYS_PARAM, [])
            if vsys == v.name:
                found = True
                assert state.name in getattr(v, self.VSYS_PARAM)
            else:
                assert state.name not in getattr(v, self.VSYS_PARAM)

        if vsys is not None:
            assert found

    def test_01_setup(self, dev, state_map):
        state = state_map.setdefault(dev)
        state.err = True
        state.fail_func = pytest.skip
        state.parent = dev
        state.name = None
        state.delete_parent = False
        state.obj = None

        if self.CLASS is None:
            pytest.skip('{0}.CLASS must be defined'.format(
                        self.__class__.__name__))
        elif self.VSYS_PARAM is None:
            pytest.skip('{0}.VSYS_PARAM must be defined'.format(
                        self.__class__.__name__))

        if isinstance(state.parent, Panorama):
            tmpl = Template(testlib.random_name())
            state.parent.add(tmpl)
            state.parent = tmpl
            state.parent.add(Vsys('vsys1'))
            state.parent.add(Vsys('vsys2'))
            state.parent.add(Vsys('vsys3'))
            state.parent.create()
            state.delete_parent = True
        else:
            vsys_list = [x.name for x in Vsys.refreshall(dev, add=False, name_only=True)]
            if not all('vsys{0}'.format(x) in vsys_list for x in range(1, 4)):
                pytest.skip('Firewall needs vsys1 - vsys3 to exist')

        args = {}
        if self.CLASS == EthernetInterface:
            state.name = testlib.get_available_interfaces(state.parent)[0]
            args = {'mode': 'layer3'}
        else:
            state.name = testlib.random_name()
        state.obj = self.CLASS(state.name, **args)
        state.parent.add(state.obj)
        state.err = False

    def test_02_create(self, dev, state_map):
        state = self.sanity(dev, state_map)

        state.err = True
        state.fail_func = pytest.xfail
        state.obj.create()
        state.err = False

    def test_03_vsys2_object_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        self.assert_imported_into(v, state)

    def test_04_update_false_does_not_update_object_rt(self, dev, state_map):
        v = 'vsys2'
        other_v = 'vsys3'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(other_v, refresh=True, update=False)

        assert isinstance(ans, Vsys)
        assert ans.name == other_v
        self.assert_imported_into(v, state)

    def test_05_import_into_lower_vsys_object_rt(self, dev, state_map):
        v = 'vsys1'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        self.assert_imported_into(v, state)

    def test_06_import_into_higher_vsys_object_rt(self, dev, state_map):
        v = 'vsys3'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        self.assert_imported_into(v, state)

    def test_07_import_into_current_vsys_object_rt(self, dev, state_map):
        v = 'vsys3'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        self.assert_imported_into(v, state)

    def test_08_unimport_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(None, refresh=True, update=True)

        assert ans is None
        self.assert_imported_into(None, state)

    def test_09_unimport_when_unimported_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(None, refresh=True, update=True)

        assert ans is None
        self.assert_imported_into(None, state)

    def test_10_vsys2_bool_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True, return_type='bool')

        assert isinstance(ans, bool)
        assert ans
        self.assert_imported_into(v, state)

    def test_11_already_in_correct_vsys_bool_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=True, return_type='bool')

        assert isinstance(ans, bool)
        assert not ans
        self.assert_imported_into(v, state)

    def test_12_same_vsys_no_update_bool_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(v, refresh=True, update=False, return_type='bool')

        assert isinstance(ans, bool)
        assert not ans
        self.assert_imported_into(v, state)

    def test_13_different_vsys_no_update_bool_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys('vsys1', refresh=True, update=False, return_type='bool')

        assert isinstance(ans, bool)
        assert ans
        self.assert_imported_into(v, state)

    def test_20_setup_for_classic_tests(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = state.obj.set_vsys(None, refresh=True, update=True)

        assert ans is None
        self.assert_imported_into(None, state)

    def test_21_classic_import_object_rt(self, dev, state_map):
        v = 'vsys1'
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(v, refresh=False, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        assert state.name in [str(x) for x in getattr(ans, self.VSYS_PARAM)]
        self.assert_imported_into(v, state)

    def test_22_classic_import_noop_object_rt(self, dev, state_map):
        v = 'vsys1'
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(v, refresh=False, update=True)

        assert isinstance(ans, Vsys)
        assert ans.name == v
        assert state.name in [str(x) for x in getattr(ans, self.VSYS_PARAM)]
        self.assert_imported_into(v, state)

    def test_23_classic_import_no_update_object_rt(self, dev, state_map):
        v = 'vsys1'
        other_v = 'vsys2'
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(other_v, refresh=False, update=False)

        assert isinstance(ans, Vsys)
        assert ans.name == other_v
        assert state.name in [str(x) for x in getattr(ans, self.VSYS_PARAM)]
        self.assert_imported_into(v, state)

    def test_24_classic_import_higher_vsys_object_rt(self, dev, state_map):
        v = 'vsys3'
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(v, refresh=False, update=True)

        previous_vsys = dev.find('vsys1')
        assert isinstance(ans, Vsys)
        assert ans.name == v
        assert state.name in [str(x) for x in getattr(ans, self.VSYS_PARAM)]
        assert previous_vsys is not None
        assert state.name not in [str(x) for x in (getattr(previous_vsys, self.VSYS_PARAM) or [])]
        self.assert_imported_into(v, state)

    def test_25_classic_import_lower_vsys_object_rt(self, dev, state_map):
        v = 'vsys2'
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(v, refresh=False, update=True)

        previous_vsys = dev.find('vsys3')
        assert isinstance(ans, Vsys)
        assert ans.name == v
        assert state.name in [str(x) for x in getattr(ans, self.VSYS_PARAM)]
        assert previous_vsys is not None
        assert state.name not in [str(x) for x in (getattr(previous_vsys, self.VSYS_PARAM) or [])]
        self.assert_imported_into(v, state)

    def test_26_classic_import_unimport_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(None, refresh=False, update=True)

        previous_vsys = dev.find('vsys2')
        assert ans is None
        assert previous_vsys is not None
        assert state.name not in (getattr(previous_vsys, self.VSYS_PARAM) or [])
        self.assert_imported_into(None, state)

    def test_27_classic_import_unimport_when_unimported_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map, True)

        ans = state.obj.set_vsys(None, refresh=False, update=True)

        assert ans is None
        self.assert_imported_into(None, state)

    def test_98_cleanup_dependencies(self, dev, state_map):
        state = state_map.setdefault(dev)

        if state.delete_parent:
            try:
                state.parent.delete()
            except Exception:
                pass
        else:
            try:
                state.obj.delete()
            except Exception:
                pass

    def test_99_removeall(self, dev, state_map):
        dev.removeall()


class TestInterfaceImport(VsysImport):
    CLASS = EthernetInterface
    VSYS_PARAM = 'interface'


class TestVirtualRouterImport(VsysImport):
    CLASS = VirtualRouter
    VSYS_PARAM = 'virtual_routers'


class TestVlanImport(VsysImport):
    CLASS = Vlan
    VSYS_PARAM = 'vlans'


class TestVirtualWireImport(VsysImport):
    CLASS = VirtualWire
    VSYS_PARAM = 'virtual_wires'


class PlaceInterface(object):
    FUNC = None
    CLASS = None
    PARAM = None

    def sanity(self, dev, state_map, fw_test=False):
        state = state_map.setdefault(dev)
        if state.err:
            state.fail_func('prereq failed')

        if fw_test:
            if not isinstance(dev, Firewall):
                pytest.skip('skipping firewall-only test')
            if self.CLASS != Zone:
                dev.vsys = None
            self.CLASS.refreshall(dev)

        if state.tmpl is None:
            dev.vsys = 'vsys2'

        return state

    def assert_placement(self, name, dev, state):
        found = False

        if state.tmpl is None and self.CLASS != Zone:
            dev.vsys = None

        obj_list = self.CLASS.refreshall(state.parent, add=False)
        for o in obj_list:
            if getattr(o, self.PARAM) is None:
                setattr(o, self.PARAM, [])
            if o.name == name:
                found = True
                assert state.name in getattr(o, self.PARAM)
            else:
                assert state.name not in getattr(o, self.PARAM)

        if name is not None:
            assert found

    def test_01_setup(self, dev, state_map):
        state = state_map.setdefault(dev)
        state.err = True
        state.fail_func = pytest.skip
        state.parent = dev
        state.tmpl = None
        state.name = None
        state.obj = None
        state.targets = [testlib.random_name() for x in range(2)]

        if self.FUNC is None:
            pytest.skip('{0}.FUNC must be defined'.format(
                        self.__class__.__name__))
        elif self.CLASS is None:
            pytest.skip('{0}.CLASS must be defined'.format(
                        self.__class__.__name__))
        elif self.PARAM is None:
            pytest.skip('{0}.PARAM must be defined'.format(
                        self.__class__.__name__))

        if isinstance(state.parent, Panorama):
            tmpl = Template(testlib.random_name())
            state.parent.add(tmpl)
            v = Vsys('vsys2')
            tmpl.add(v)
            state.parent = v
            tmpl.create()
            state.tmpl = tmpl
        else:
            vsys_list = [x.name for x in Vsys.refreshall(dev, add=False, name_only=True)]
            if 'vsys2' not in vsys_list:
                pytest.skip('Firewall needs vsys2 to exist')

        cls_args = {}
        eth_args = {'mode': 'layer3'}
        if self.CLASS == Vlan:
            eth_args['mode'] = 'layer2'
        elif self.CLASS == Zone:
            cls_args['mode'] = 'layer3'

        state.name = testlib.get_available_interfaces(state.parent)[0]
        state.obj = EthernetInterface(state.name, **eth_args)
        state.parent.add(state.obj)

        if state.tmpl is None:
            dev.vsys = 'vsys2'

        instances = [self.CLASS(state.targets[x], **cls_args) for x in range(2)]
        for x in instances:
            state.parent.add(x)
            x.create()

        state.err = False

    def test_02_create(self, dev, state_map):
        state = self.sanity(dev, state_map)

        state.err = True
        state.fail_func = pytest.xfail
        state.obj.create()
        state.obj.set_vsys('vsys2', refresh=True, update=True)
        state.err = False

    def test_03_set_no_priors_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        self.assert_placement(state.targets[i], dev, state)

    def test_04_update_false_does_not_update_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[1], refresh=True, update=False)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[1]
        self.assert_placement(state.targets[0], dev, state)

    def test_05_change_forward_object_rt(self, dev, state_map):
        i = 1
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        self.assert_placement(state.targets[i], dev, state)

    def test_06_change_backward_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        self.assert_placement(state.targets[i], dev, state)

    def test_07_change_to_current_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        self.assert_placement(state.targets[i], dev, state)

    def test_08_remove_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(None, refresh=True, update=True)

        assert ans is None
        self.assert_placement(None, dev, state)

    def test_09_remove_when_not_present_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(None, refresh=True, update=True)

        assert ans is None
        self.assert_placement(None, dev, state)

    def test_10_change_bool_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True, return_type='bool')

        assert isinstance(ans, bool)
        assert ans
        self.assert_placement(state.targets[i], dev, state)

    def test_11_in_place_bool_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=True, return_type='bool')

        assert isinstance(ans, bool)
        assert not ans
        self.assert_placement(state.targets[i], dev, state)

    def test_12_same_spot_no_update_bool_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=True, update=False, return_type='bool')

        assert isinstance(ans, bool)
        assert not ans
        self.assert_placement(state.targets[i], dev, state)

    def test_13_different_spot_no_update_bool_rt(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(state.targets[1], refresh=True, update=False, return_type='bool')

        assert isinstance(ans, bool)
        assert ans
        self.assert_placement(state.targets[0], dev, state)

    def test_20_setup_for_classic_tests(self, dev, state_map):
        state = self.sanity(dev, state_map)

        ans = getattr(state.obj, self.FUNC)(None, refresh=True, update=True)

        assert ans is None
        self.assert_placement(None, dev, state)

    def test_21_classic_placement_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=False, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        assert state.name in [str(x) for x in getattr(ans, self.PARAM)]
        self.assert_placement(state.targets[i], dev, state)

    def test_22_classic_noop_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=False, update=True)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        assert state.name in [str(x) for x in getattr(ans, self.PARAM)]
        self.assert_placement(state.targets[i], dev, state)

    def test_23_classic_no_update_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(state.targets[1], refresh=False, update=False)

        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[1]
        assert state.name in [str(x) for x in getattr(ans, self.PARAM)]
        self.assert_placement(state.targets[0], dev, state)

    def test_24_classic_placement_move_forward_object_rt(self, dev, state_map):
        i = 1
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=False, update=True)

        prev_obj = state.parent.find(state.targets[0])
        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        assert state.name in [str(x) for x in getattr(ans, self.PARAM)]
        assert prev_obj is not None
        assert state.name not in [str(x) for x in (getattr(prev_obj, self.PARAM) or [])]
        self.assert_placement(state.targets[i], dev, state)

    def test_25_classic_placement_move_backward_object_rt(self, dev, state_map):
        i = 0
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(state.targets[i], refresh=False, update=True)

        prev_obj = state.parent.find(state.targets[1])
        assert isinstance(ans, self.CLASS)
        assert ans.name == state.targets[i]
        assert state.name in [str(x) for x in getattr(ans, self.PARAM)]
        assert prev_obj is not None
        assert state.name not in [str(x) for x in (getattr(prev_obj, self.PARAM) or [])]
        self.assert_placement(state.targets[i], dev, state)

    def test_26_classic_remove_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(None, refresh=False, update=True)

        prev_obj = state.parent.find(state.targets[0])
        assert ans is None
        assert prev_obj is not None
        assert state.name not in [str(x) for x in (getattr(prev_obj, self.PARAM) or [])]
        self.assert_placement(None, dev, state)

    def test_27_classic_remove_when_already_removed_object_rt(self, dev, state_map):
        state = self.sanity(dev, state_map, True)

        ans = getattr(state.obj, self.FUNC)(None, refresh=False, update=True)

        assert ans is None
        self.assert_placement(None, dev, state)

    def test_98_cleanup_dependencies(self, dev, state_map):
        state = self.sanity(dev, state_map)

        if state.tmpl is not None:
            try:
                state.tmpl.delete()
            except Exception:
                pass
        else:
            if self.CLASS != Zone:
                dev.vsys = None
            instances = self.CLASS.refreshall(dev, name_only=True)
            dev.vsys = 'vsys2'
            for o in instances:
                if o.name in state.targets:
                    try:
                        o.delete()
                    except Exception:
                        pass
            try:
                state.obj.delete()
            except Exception:
                pass
            if dev.vsys is not None:
                dev.vsys = None

    def test_99_removeall(self, dev, state_map):
        dev.removeall()


class TestVirtualRouterPlacement(PlaceInterface):
    FUNC = 'set_virtual_router'
    CLASS = VirtualRouter
    PARAM = 'interface'


class TestVlanPlacement(PlaceInterface):
    FUNC = 'set_vlan'
    CLASS = Vlan
    PARAM = 'interface'


class TestZonePlacement(PlaceInterface):
    FUNC = 'set_zone'
    CLASS = Zone
    PARAM = 'interface'



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
try:
    from unittest import mock
except ImportError:
    import mock
import unittest

import pandevice.firewall
import pandevice.panorama


class TestUserId(unittest.TestCase):
    """
    [Test section:  userid]

    Verify some userid methods that can be tested without
    a live device
    """

    def test_login(self):
        expected = b'<uid-message><version>1.0</version>' \
                   b'<type>update</type><payload><login>' \
                   b'<entry ip="10.1.1.1" name="example.com\username" timeout="10" />' \
                   b'</login></payload></uid-message>'
        vsys = 'vsys3'

        fw = pandevice.firewall.Firewall(
            'fw1', 'user', 'passwd', 'authkey', serial='Serial', vsys=vsys)
        fw.xapi
        fw._xapi_private.user_id = mock.Mock()

        fw.userid.login(r'example.com\username', '10.1.1.1', timeout=10)

        fw._xapi_private.user_id.assert_called_once_with(cmd=expected, vsys=vsys)


if __name__=='__main__':
    unittest.main()



# Copyright (c) 2014, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
try:
    from unittest import mock
except ImportError:
    import mock
import unittest
import pan.xapi

import pandevice.base as Base
import pandevice.device
import pandevice.errors as Err
import pandevice.firewall
import pandevice.ha
import pandevice.network
import pandevice.objects
import pandevice.panorama

import xml.etree.ElementTree as ET


class TestTemplates(unittest.TestCase):
    def setUp(self):
        self.pano = pandevice.panorama.Panorama(
            'foo', 'bar', 'baz', 'apikey')
        self.pano._version_info = (8, 0, 0)

    def test_template_element_str_stays_consistent(self):
        expected = ''.join([
            '<entry name="blah"><description>my description</description>',
            '<settings><default-vsys>vsys1</default-vsys></settings>',
            '<config><devices><entry name="localhost.localdomain"><vsys>',
            '<entry name="vsys1"><import><network><interface /></network>',
            '</import></entry></vsys><network><virtual-router>',
            '<entry name="some vr"><admin-dists><static>42</static>',
            '<ebgp>21</ebgp></admin-dists></entry></virtual-router>',
            '</network></entry></devices></config></entry>',
        ])

        xml_tree = ET.fromstring('<result>{0}</result>'.format(expected))
        t = pandevice.panorama.Template()
        self.pano.add(t)
        other = t.refreshall_from_xml(xml_tree)[0]

        self.assertEqual(expected, other.element_str().decode('utf-8'))


class TestNearestPandevice(unittest.TestCase):
    """
    [Test section: nearest_pandevice()]

    Verify that `nearest_pandevice()` returns the correct PanObject.

    All tests have a setup of the following:
        Panorama > DeviceGroup > Firewall > AddressObject

    In this configuration, running .nearest_pandevice() should return the
    panorama device in all cases but running it from the AddressObject.
    """
    def setUp(self):
        self.panorama = pandevice.panorama.Panorama(
            'foo', 'bar', 'baz', 'apikey')
        self.device_group = pandevice.panorama.DeviceGroup(
            'My pandevice Group')
        self.firewall = pandevice.firewall.Firewall(
            'foo', 'bar', 'baz', 'apikey')
        self.address_object = pandevice.objects.AddressObject(
            'webserver', '192.168.1.100', description='Intranet web server',
            tag=['http', 'https'])

        self.firewall.add(self.address_object)
        self.device_group.add(self.firewall)
        self.panorama.add(self.device_group)

        self.assertEqual(self.firewall, self.address_object.parent)
        self.assertEqual([], self.address_object.children)
        self.assertEqual(self.device_group, self.firewall.parent)
        self.assertEqual([self.address_object, ], self.firewall.children)
        self.assertEqual(self.panorama, self.device_group.parent)
        self.assertEqual([self.firewall, ], self.device_group.children)
        self.assertEqual(None, self.panorama.parent)
        self.assertEqual([self.device_group, ], self.panorama.children)

    def test_nearest_pandevice_from_addressobject_in_pano_dg_fw_ao_chain(self):
        """Runs nearest_pandevice() on the AddressObject.

        This should return the Firewall PanDevice.
        """
        ret_val = self.address_object.nearest_pandevice()

        self.assertEqual(self.firewall, ret_val)

    def test_nearest_pandevice_from_firewall_in_pano_dg_fw_ao_chain(self):
        """Runs nearest_pandevice() on the firewall.

        This should return the Panorama PanDevice.
        """
        ret_val = self.firewall.nearest_pandevice()

        self.assertEqual(self.panorama, ret_val)

    def test_nearest_pandevice_from_device_group_in_pano_dg_fw_ao_chain(self):
        """Runs nearest_pandevice() on the device group.

        This should return the Panorama PanDevice.
        """
        ret_val = self.device_group.nearest_pandevice()

        self.assertEqual(self.panorama, ret_val)

    def test_nearest_pandevice_from_panorama_in_pano_dg_fw_ao_chain(self):
        """Runs nearest_pandevice() on the panorama.

        This should return the Panorama PanDevice.
        """
        ret_val = self.panorama.nearest_pandevice()

        self.assertEqual(self.panorama, ret_val)

    def test_nearest_pandevice_from_firewall_with_no_parents_returns_self(self):
        fw = pandevice.firewall.Firewall(
            'foo', 'bar', 'baz', 'apikey')

        ret_val = fw.nearest_pandevice()

        self.assertEqual(fw, ret_val)


class TestElementStr_7_0(unittest.TestCase):
    """
    [Test section:  element_str()]

    Verify the XML created under various circumstances for various
    objects and setups:

        * HighAvailability with HA1 and HA2 children
        * VirtualRouter with StaticRoute child
        * EthernetInterface
        * Firewall
        * AddressObject
    """

    # 1) HighAvailability with HA1 and HA2 children
    def test_element_str_from_highavailability_with_ha1_and_ha2_children(self):
        expected = b''.join([
            b'<high-availability><enabled>yes</enabled><group><entry name="1">',
            b'<description>my ha conf description</description>',
            b'<configuration-synchronization><enabled>yes</enabled>',
            b'</configuration-synchronization><peer-ip>10.5.1.5</peer-ip>',
            b'<mode><active-passive><passive-link-state>passive state',
            b'</passive-link-state></active-passive></mode>',
            b'<state-synchronization><enabled>no</enabled><ha2-keep-alive>',
            b'<enabled>yes</enabled><action>ha2 do stuff</action><threshold>',
            b'2</threshold></ha2-keep-alive></state-synchronization></entry>',
            b'</group><interface><ha1><ip-address>10.5.1.1</ip-address>',
            b'<netmask>255.255.255.0</netmask><port>ethernet1/6</port>',
            b'<gateway>10.5.1.2</gateway><link-speed>1000</link-speed>',
            b'<link-duplex>auto</link-duplex><monitor-hold-time>7',
            b'</monitor-hold-time></ha1><ha1-backup /><ha2>',
            b'<ip-address>10.6.1.1</ip-address><netmask>255.255.255.0',
            b'</netmask><port>ethernet1/7</port><gateway>10.6.1.2</gateway>',
            b'<link-speed>1000</link-speed><link-duplex>auto</link-duplex>',
            b'</ha2><ha2-backup /><ha3 /></interface></high-availability>',
        ])

        h1o = pandevice.ha.HA1(
            name='ha101', ip_address='10.5.1.1', netmask='255.255.255.0',
            port='ethernet1/6', gateway='10.5.1.2', link_speed='1000',
            link_duplex='auto', monitor_hold_time=7)
        h2o = pandevice.ha.HA2(
            name='ha202', ip_address='10.6.1.1', netmask='255.255.255.0',
            port='ethernet1/7', gateway='10.6.1.2', link_speed='1000',
            link_duplex='auto')
        ha_config = pandevice.ha.HighAvailability(
            'my high availability config', True, '1', 'my ha conf description',
            True, '10.5.1.5', 'active-passive', 'passive state', False, True,
            'ha2 do stuff', 2)

        ha_config.add(h1o)
        ha_config.add(h2o)
        ha_config.retrieve_panos_version = mock.Mock(return_value=(7, 0, 0))

        ret_val = ha_config.element_str()

        self.assertEqual(expected, ret_val,
            '\n{0}\n{1}'.format(expected, ret_val))

    # 2) VirtualRouter with StaticRoute child
    def test_element_str_from_virtualrouter_with_sr_parent(self):
        '''StaticRoute > VirtualRouter'''
        expected = b''.join([
            b'<entry name="default"><interface><member>ethernet1/3</member>',
            b'</interface><routing-table><ip><static-route>',
            b'<entry name="my static route"><destination>0.0.0.0/0',
            b'</destination><nexthop><ip-address>192.168.5.1</ip-address>',
            b'</nexthop><interface>ethernet1/4</interface><metric>10</metric>',
            b'</entry></static-route></ip></routing-table></entry>',
        ])

        vro = pandevice.network.VirtualRouter('default', 'ethernet1/3')
        sro = pandevice.network.StaticRoute(
            'my static route', '0.0.0.0/0', 'ip-address',
            '192.168.5.1', 'ethernet1/4')

        vro.add(sro)

        ret_val = vro.element_str()

        self.assertEqual(expected, ret_val)

    # 3) EthernetInterface
    def test_element_str_from_ethernetinterface(self):
        expected = b''.join([
            b'<entry name="ethernet1/1"><layer3><ip><entry name="10.1.1.1" />',
            b'</ip></layer3><link-speed>1000</link-speed><link-duplex>auto',
            b'</link-duplex><link-state>auto</link-state></entry>',
        ])

        o = pandevice.network.EthernetInterface(
            'ethernet1/1', 'layer3', '10.1.1.1', link_speed='1000',
            link_duplex='auto', link_state='auto')

        o_str = o.element_str()
        self.assertEqual(expected, o_str)

    def test_element_str_from_ethernetinterface_in_en_l3s_arp(self):
        '''EthernetInterface > Layer3Subinterface > Arp'''
        expected = b''.join([
            b'<entry name="ethernet1/1"><layer3><ip>',
            b'<entry name="10.3.6.12" /></ip><units>',
            b'<entry name="ethernet1/1.355"><tag>355</tag><ip>',
            b'<entry name="10.20.30.40/24" /></ip><mtu>1500</mtu>',
            b'<adjust-tcp-mss>yes</adjust-tcp-mss><arp>',
            b'<entry name="10.5.10.15"><hw-address>00:30:48:52:cd:dc',
            b'</hw-address></entry></arp></entry></units></layer3></entry>',
        ])

        ao = pandevice.network.Arp('10.5.10.15', '00:30:48:52:cd:dc')
        l3so = pandevice.network.Layer3Subinterface(
            'ethernet1/1.355', 355, '10.20.30.40/24',
            mtu=1500, adjust_tcp_mss=True)
        eio = pandevice.network.EthernetInterface(
            'ethernet1/1', mode='layer3', ip='10.3.6.12')

        l3so.add(ao)
        eio.add(l3so)

        # This is actually a 7.x test
        for o in (ao, l3so, eio):
            o.retrieve_panos_version = mock.Mock(return_value=(7, 0, 0))

        ret_val = eio.element_str()

        self.assertEqual(expected, ret_val)

    def test_element_str_from_ethernetinterface_for_aggregate_group(self):
        expected = b''.join([
            b'<entry name="ethernet1/1"><aggregate-group>ae1',
            b'</aggregate-group></entry>',
        ])
        eio = pandevice.network.EthernetInterface(
            'ethernet1/1', 'aggregate-group', '10.3.6.12',
            aggregate_group='ae1')

        ret_val = eio.element_str()

        self.assertEqual(expected, ret_val)

    # 4) Firewall
    """
    * vsys of the firewall
    * multi_vsys variable is
    * serial is set / not set
    * parent is pano / device group
    """
    def test_element_str_from_firewall_with_pano_parent_and_systemsettings_child(self):
        expected = b''.join([
            b'<entry name="Serial"><vsys>',
            b'<entry name="vsys1" /></vsys></entry>',
        ])

        fw = pandevice.firewall.Firewall(
            'fw1', 'user', 'passwd', 'authkey', serial='Serial', vsys='vsys3')
        pano = pandevice.panorama.Panorama('10.100.5.2')
        conf = pandevice.device.SystemSettings(
            hostname='Hostname-Setting',
            domain='paloaltonetworks.com',
            ip_address='10.20.30.40',
            netmask='255.255.255.0',
            default_gateway='10.20.30.1',
            panorama='10.100.5.2',
            login_banner="This is not the firewall you're looking for..",
            update_server='8.8.8.8',
        )

        fw.add(conf)
        pano.add(fw)

        ret_val = fw.element_str()

        self.assertEqual(expected, ret_val)

    def test_element_str_from_firewall_without_serial_number_raises_error(self):
        fw = pandevice.firewall.Firewall('foo')

        self.assertRaises(
            ValueError,
            fw.element_str)

    def test_element_str_from_firewall_with_dg_pano_parents_and_multi_vsys(self):
        expected = b''.join([
            b'<entry name="serial"><vsys><entry name="vsys3" />',
            b'</vsys></entry>',
        ])

        fw = pandevice.firewall.Firewall(
            'fw1', 'user', 'passwd', 'authkey',
            serial='serial', vsys='vsys3', multi_vsys=True)
        dg = pandevice.panorama.DeviceGroup('my group')
        p = pandevice.panorama.Panorama('pano')

        dg.add(fw)
        p.add(dg)

        ret_val = fw.element_str()

        self.assertEqual(expected, ret_val)

    # 5) AddressObject
    def test_element_str_from_addressobject(self):
        expected = b''.join([
            b'<entry name="webserver"><ip-netmask>192.168.1.100</ip-netmask>',
            b'<description>Intranet web server</description><tag><member>',
            b'https</member><member>http</member></tag></entry>',
        ])
        o = pandevice.objects.AddressObject(
            'webserver', '192.168.1.100', description='Intranet web server',
            tag=['https', 'http'])

        o_str = o.element_str()
        self.assertEqual(expected, o_str)


class TestXpaths_7_0(unittest.TestCase):
    """
    [Test section: xpaths]
    Test that both set and edit xpaths are reported correctly.

    These test cases have various object constructions, and verify that both
    the 'set' and 'edit/delete' xpath can correctly be identified.
    """

    # HA1 / HA2 Tests
    def test_edit_xpath_from_ha1_with_ha_fw_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/deviceconfig/high-availability",
            "/interface/ha1",
        ])

        child = pandevice.ha.HA1()
        parent = pandevice.ha.HighAvailability()
        fw = pandevice.firewall.Firewall()

        parent.add(child)
        fw.add(parent)

        ret_val = child.xpath()

        self.assertEqual(expected, ret_val)

    def test_edit_xpath_from_ha1_with_ha_fw_pano_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/deviceconfig/high-availability",
            "/interface/ha1",
        ])

        child = pandevice.ha.HA1('ha1')
        parent = pandevice.ha.HighAvailability('ha parent')
        fw = pandevice.firewall.Firewall('myfw')
        pano = pandevice.panorama.Panorama('panorama')

        parent.add(child)
        fw.add(parent)
        pano.add(fw)

        ret_val = child.xpath()

        self.assertEqual(expected, ret_val)

    def test_edit_xapth_from_ha2_with_ha_fw_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/deviceconfig/high-availability",
            "/interface/ha2",
        ])

        child = pandevice.ha.HA2('child')
        parent = pandevice.ha.HighAvailability('parent')
        fw = pandevice.firewall.Firewall('myfw')

        parent.add(child)
        fw.add(parent)

        ret_val = child.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_ha2_with_ha_fw_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/deviceconfig/high-availability",
            "/interface",
        ])

        child = pandevice.ha.HA2('child')
        parent = pandevice.ha.HighAvailability('HighAvail')
        fw = pandevice.firewall.Firewall('myfw')

        parent.add(child)
        fw.add(parent)

        ret_val = child.xpath_short()

        self.assertEqual(ret_val, expected)
        self.assertEqual(expected, ret_val)

    # VirtualRouter tests
    def test_edit_xpath_from_virtualrouter_with_sr_fw_parents(self):
        '''Firewall > VirtualRouter > StaticRoute'''
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/routing-table/ip/static-route/entry[@name='sr']",
            "/network/virtual-router/entry[@name='vr']",
        ])

        child = pandevice.network.VirtualRouter('vr')
        parent = pandevice.network.StaticRoute('sr')
        fw = pandevice.firewall.Firewall('fw')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        parent.add(child)
        fw.add(parent)

        ret_val = child.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_virtualrouter_with_sr_fw_parents(self):
        '''Firewall > VirtualRouter > StaticRoute'''
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/routing-table/ip/static-route/entry[@name='sr']",
            "/network/virtual-router",
        ])

        child = pandevice.network.VirtualRouter('vr')
        parent = pandevice.network.StaticRoute('sr')
        fw = pandevice.firewall.Firewall('fw')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        parent.add(child)
        fw.add(parent)

        ret_val = child.xpath_short()

        self.assertEqual(expected, ret_val)

    # Arp (EthernetInterface) tests
    def test_edit_xpath_from_arp_with_l3s_ei_fw_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/network/interface/ethernet/entry[@name='Eth Interface Object']",
            "/layer3/units/entry[@name='Layer3 Subint Object']",
            "/arp/entry[@name='arp object']",
        ])

        ao = pandevice.network.Arp('arp object')
        l3so = pandevice.network.Layer3Subinterface('Layer3 Subint Object')
        eio = pandevice.network.EthernetInterface('Eth Interface Object')
        fw = pandevice.firewall.Firewall('fw')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        l3so.add(ao)
        eio.add(l3so)
        fw.add(eio)

        ret_val = ao.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_arp_with_l3s_ei_fw_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/network/interface/ethernet/entry[@name='Eth Interface Object']",
            "/layer3/units/entry[@name='Layer3 Subint Object']",
            "/arp",
        ])

        ao = pandevice.network.Arp('arp object')
        l3so = pandevice.network.Layer3Subinterface('Layer3 Subint Object')
        eio = pandevice.network.EthernetInterface('Eth Interface Object')
        fw = pandevice.firewall.Firewall('fw')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        l3so.add(ao)
        eio.add(l3so)
        fw.add(eio)

        ret_val = ao.xpath_short()

        self.assertEqual(expected, ret_val)

    # Firewall tests
    def test_edit_xpath_from_firewall(self):
        # This is not a valid xpath, but its what should happen
        # if there is no parent
        expected = ''.join([
            "/devices/entry[@name='serial']",
        ])

        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        ret_val = fw.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_firewall(self):
        # This is not a valid xpath, but its what should happen
        # if there is no parent
        expected = ''.join([
            "/devices",
        ])

        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        ret_val = fw.xpath_short()

        self.assertEqual(expected, ret_val)

    def test_edit_xpath_from_firewall_with_pano_parent(self):
        expected = "/config/mgt-config/devices/entry[@name='serial']"

        p = pandevice.panorama.Panorama('pano')
        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        p.add(fw)

        ret_val = fw.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_firewall_with_pano_parent(self):
        expected = "/config/mgt-config/devices"

        p = pandevice.panorama.Panorama('pano')
        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        p.add(fw)

        ret_val = fw.xpath_short()

        self.assertEqual(expected, ret_val)

    def test_edit_xpath_from_firewall_with_dg_pano_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/device-group/entry[@name='my group']/devices",
            "/entry[@name='serial']",
        ])

        p = pandevice.panorama.Panorama('pano')
        dg = pandevice.panorama.DeviceGroup('my group')
        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        dg.add(fw)
        p.add(dg)

        ret_val = fw.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_firewall_with_dg_pano_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/device-group/entry[@name='my group']/devices",
        ])

        p = pandevice.panorama.Panorama('pano')
        dg = pandevice.panorama.DeviceGroup('my group')
        fw = pandevice.firewall.Firewall(
            'foo', vsys='vsys2', serial='serial')

        dg.add(fw)
        p.add(dg)

        ret_val = fw.xpath_short()

        self.assertEqual(expected, ret_val)

    # AddressObject tests
    def test_edit_xpath_from_addressobject_with_fw_parent(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/vsys/entry[@name='vsys2']",
            "/address/entry[@name='ntp server']",
        ])

        ao = pandevice.objects.AddressObject('ntp server')
        fw = pandevice.firewall.Firewall('fw', vsys='vsys2')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        fw.add(ao)

        ret_val = ao.xpath()

        self.assertEqual(expected, ret_val)

    def test_edit_xpath_from_addressobject_with_dg_panorama_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/device-group/entry[@name='My Group']",
            "/address/entry[@name='webproxy']",
        ])

        ao = pandevice.objects.AddressObject('webproxy')
        dg = pandevice.panorama.DeviceGroup('My Group')
        pano = pandevice.panorama.Panorama('My Panorama')
        pano.get_device_version = mock.Mock(return_value=(7, 0, 0))

        dg.add(ao)
        pano.add(dg)

        ret_val = ao.xpath()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_addressobject_with_fw_parent(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/vsys/entry[@name='vsys2']",
            "/address",
        ])

        ao = pandevice.objects.AddressObject('ntp server')
        fw = pandevice.firewall.Firewall('fw', vsys='vsys2')
        fw.get_device_version = mock.Mock(return_value=(7, 0, 0))

        fw.add(ao)

        ret_val = ao.xpath_short()

        self.assertEqual(expected, ret_val)

    def test_set_xpath_from_addressobject_with_dg_panorama_parents(self):
        expected = ''.join([
            "/config/devices/entry[@name='localhost.localdomain']",
            "/device-group/entry[@name='My Group']",
            "/address",
        ])

        ao = pandevice.objects.AddressObject('webproxy')
        dg = pandevice.panorama.DeviceGroup('My Group')
        pano = pandevice.panorama.Panorama('My Panorama')
        pano.get_device_version = mock.Mock(return_value=(7, 0, 0))

        dg.add(ao)
        pano.add(dg)

        ret_val = ao.xpath_short()

        self.assertEqual(expected, ret_val)

    def test_xpath_from_addressobject_with_pano_parent(self):
        expected = "/config/shared/address/entry[@name='shared ao']"

        ao = pandevice.objects.AddressObject('shared ao')
        pano = pandevice.panorama.Panorama('pano')
        pano.get_device_version = mock.Mock(return_value=(7, 0, 0))

        pano.add(ao)

        ret_val = ao.xpath()

        self.assertEqual(expected, ret_val)


class TestVariousSubinterfaceXpaths(unittest.TestCase):
    def test_l2_subinterface_with_firewall_parent(self):
        fw = pandevice.firewall.Firewall('192.168.1.1', 'admin', 'admin', vsys='vsys2')
        iface = pandevice.network.EthernetInterface('ethernet1/3', 'layer2')
        eth = pandevice.network.Layer2Subinterface('ethernet1/3.3', 3)
        iface.add(eth)
        fw.add(iface)

        expected = eth.xpath()

        fw.add(eth)

        self.assertEqual(expected, eth.xpath())

    def test_l2_subinterface_with_vsys_parent(self):
        fw = pandevice.firewall.Firewall('192.168.1.1', 'admin', 'admin')
        vsys = pandevice.device.Vsys('vsys2')
        iface = pandevice.network.EthernetInterface('ethernet1/3', 'layer2')
        eth = pandevice.network.Layer2Subinterface('ethernet1/3.3', 3)
        iface.add(eth)
        vsys.add(iface)
        fw.add(vsys)

        expected = eth.xpath()

        vsys.add(eth)

        self.assertEqual(expected, eth.xpath())

    def test_l3_subinterface_with_firewall_parent(self):
        fw = pandevice.firewall.Firewall('192.168.1.1', 'admin', 'admin', vsys='vsys3')
        iface = pandevice.network.EthernetInterface('ethernet1/4', 'layer3')
        eth = pandevice.network.Layer3Subinterface('ethernet1/4.4', 4)
        iface.add(eth)
        fw.add(iface)

        expected = eth.xpath()

        fw.add(eth)

        self.assertEqual(expected, eth.xpath())

    def test_l3_subinterface_with_vsys_parent(self):
        fw = pandevice.firewall.Firewall('192.168.1.1', 'admin', 'admin')
        vsys = pandevice.device.Vsys('vsys3')
        iface = pandevice.network.EthernetInterface('ethernet1/4', 'layer3')
        eth = pandevice.network.Layer2Subinterface('ethernet1/4.4', 4)
        iface.add(eth)
        vsys.add(iface)
        fw.add(vsys)

        expected = eth.xpath()

        vsys.add(eth)

        self.assertEqual(expected, eth.xpath())


if __name__=='__main__':
    unittest.main()
